/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Clippings.
 *
 * The Initial Developer of the Original Code is 
 * Alex Eng <ateng@users.sourceforge.net>.
 * Portions created by the Initial Developer are Copyright (C) 2005-2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */

//
// Integration with host application
//

var gClippings = {
  // Constants for HTML paste options
  HTMLPASTE_ASK_THE_USER: 0,
  HTMLPASTE_AS_HTML:      1,
  HTMLPASTE_AS_IS:        2,

  dataSrcInitialized:    false,
  showDialog:            true,
  appUtils:              null,
  clippingsSvc:          null,
  strBundle:             null,
  _clippingsListener:    null,
  _clippingsToolbar:     null,
  _isErrMenuItemVisible: false,
  _clippingSubst:        null,
  _ds:                   null,

  statusBarObserver: {
    onDragOver: function (aEvent, aFlavour, aSession)
    {
      aEvent.target.setAttribute("ae-dnd", "true");
    },

    onDragExit: function (aEvent, aSession)
    {
      aEvent.target.removeAttribute("ae-dnd");
    },

    onDrop: function (aEvent, aTransferData, aSession)
    {
      if (! gClippings.dataSrcInitialized) {
	// The RDF data source has to be initialized if it has not already been
	// done so, otherwise RDF node creation will fail, and the new
	// Clippings entry will never be created.
	// This initialization is done in the code for the Clippings popup
	// menu's `onpopupshowing' event handler.
	gClippings.initClippingsPopup(document.getElementById("ae-clippings-popup-1"), document.getElementById("ae-clippings-menu-1"));
      }

      var text = aTransferData.data;
      var result = gClippings.clippingsSvc.newClippingFromText(text, gClippings.showDialog, window, null, false);
      if (result) {
	window.setTimeout("gClippings.saveClippings();", 100);
      }
    },

    getSupportedFlavours: function ()
    {
      var flavours = new FlavourSet();
      flavours.appendFlavour("text/unicode");
      return flavours;
    }
  },


  // Method handleEvent() effectively makes gClippings an implementation of
  // the EventListener interface; therefore it can be passed as the listener
  // argument to window.addEventListener() and window.removeEventListener()

  handleEvent: function (aEvent)
  {
    // When this method is invoked, 'this' will not refer to gClippings
    var _this = gClippings;

    if (aEvent.type == "load") {
      _this.initClippings();
    }
    else if (aEvent.type == "unload") {
      _this.unload();
      window.removeEventListener("load", _this, false);
      window.removeEventListener("unload", _this, false);
      window.removeEventListener("focus", _this, false);

      var hostAppCxtMenu = document.getElementById("contentAreaContextMenu");
      hostAppCxtMenu.removeEventListener("popupshowing", 
					 _this._fnInitContextMenuItem,
					 false);
      hostAppCxtMenu.removeEventListener("popuphiding",
					 _this._fnHideContextMenuItem,
					 false);
    }
    else if (aEvent.type == "focus") {
      _this.focusHostAppWnd();
    }
  },


  //
  // Methods invoked by overlay code
  //

  alert: function (aMessage)
  {
    var title = this.strBundle.getString('appName');
    this.appUtils.alertEx(title, aMessage);
  },


  newFromClipboard: function () 
  {
    if (! this.dataSrcInitialized) {
      // The RDF data source has to be initialized if it has not already been
      // done so, otherwise RDF node creation will fail, and the new Clippings
      // entry will never be created.
      // This initialization is done in the code for the Clippings popup menu's
      // `onpopupshowing' event handler.
      this.initClippingsPopup(document.getElementById("ae-clippings-popup-1"),
			      document.getElementById("ae-clippings-menu-1"));
    }

    var txt = this.appUtils.getTextFromClipboard();
    if (! txt) {
      this.alert(this.strBundle.getString('errorNoClipboardContent'));
      return;
    }

    var result = this.clippingsSvc.newClippingFromText(txt, this.showDialog, window, null, false);
    if (result) {
      window.setTimeout("gClippings.saveClippings();", 1);
    }
  },


  newFromTextbox: function ()
  {
    if (! this.dataSrcInitialized) {
      this.initClippingsPopup(document.getElementById("ae-clippings-popup-1"),
			      document.getElementById("ae-clippings-menu-1"));
    }

    var cxtMenu = document.getElementById("contentAreaContextMenu");
    var result;

    // ** FIREFOX 2 **
    // gContextMenu.onTextInput is also true inside a rich edit box!
    if (gContextMenu.onTextInput) {
      // Must explicitly close the browser content area context menu -
      // otherwise it will reappear while the New Clipping dialog is open if
      // the Clippings submenu needs to be rebuilt.
      cxtMenu.hidePopup();

      // Get the node where the context menu is invoked from.
      var textbox = document.popupNode;
      if (textbox instanceof HTMLInputElement || textbox instanceof HTMLTextAreaElement) {
	var text;
	if (textbox.selectionStart == textbox.selectionEnd) {
	  text = textbox.value;
	}
	else {
	  text = textbox.value.substring(textbox.selectionStart, 
					 textbox.selectionEnd);
	}
	result = this.clippingsSvc.newClippingFromText(text, this.showDialog, window, null, false);
      }
      else {
	var doc = document.popupNode.ownerDocument;

	// ** FIREFOX 2 **  Rich edit field
	if (this.appUtils.getHostAppVersionStr().split(".")[0] >= 2 && doc) {
	  cxtMenu.hidePopup();
	  var plainText = doc.defaultView.getSelection();

	  // New (from entire contents)
	  if (plainText == "") {
	    goDoCommand('cmd_selectAll');
	    plainText = doc.defaultView.getSelection();
	    var isTextSelected = true;
	  }
	  result = this.clippingsSvc.newClippingFromText(plainText, this.showDialog, window, null, false);
	  if (isTextSelected) {
	    // TO DO: Deselect text, leaving the cursor where it was.
	  }
	}
	else {
	  this.appUtils.log("In gClippings.newFromTextbox(): textbox = " 
			    + textbox);
	}
      }
    }
    // ** FIREFOX 1.x ONLY **  Rich edit field
    else if (gContextMenu.inFrame) {
      var doc = document.popupNode.ownerDocument;
      cxtMenu.hidePopup();

      if (doc.designMode == "on") {
	var plainText = doc.defaultView.getSelection();

	// New (from entire contents)
	if (plainText == "") {
	  goDoCommand('cmd_selectAll');
	  plainText = doc.defaultView.getSelection();
	  var isTextSelected = true;
	}
      	result = this.clippingsSvc.newClippingFromText(plainText, this.showDialog, window, null, false);
	if (isTextSelected) {
	  // TO DO: Deselect text, leaving the cursor where it was.
	}
      }
    }

    if (result) {
      window.setTimeout("gClippings.saveClippings();", 1);
    }
  },


  extNewClipping: function ()
  {
    // This function is meant to be invoked by the Clippings toolbar
    var cmdCopy = document.getElementById("cmd_copy");
    if (cmdCopy.getAttribute("disabled")) {
      this.appUtils.beep();
      return;
    }

    goDoCommand("cmd_copy");
    var selection = this.appUtils.getTextFromClipboard();

    if (selection) {
      var result = this.clippingsSvc.newClippingFromText(selection, this.showDialog, window, null, false);
      if (result) {
	window.setTimeout("gClippings.saveClippings();", 1);
      }
    }
    else {
      var strBundle = document.getElementById("ae-clippings-strings");
      this.alert(this.strBundle.getString('errorNoSelection'));
    }
  },


  newFromSelection: function ()
  {
    if (! this.dataSrcInitialized) {
      this.initClippingsPopup(document.getElementById("ae-clippings-popup-1"),
			      document.getElementById("ae-clippings-menu-1"));
    }
    var focusedWnd = document.commandDispatcher.focusedWindow;
    var selection = focusedWnd.getSelection();
    if (selection) {
      var result = this.clippingsSvc.newClippingFromText(selection, this.showDialog, window, null, false);
      if (result) {
	window.setTimeout("gClippings.saveClippings();", 1);
      }
    }
    else {
      var strBundle = document.getElementById("ae-clippings-strings");
      this.alert(this.strBundle.getString('errorNoSelection'));
    }
  },


  openClippingsManager: function () 
  {
    var wnd = window.open("chrome://clippings/content/clippings.xul",
			  "clippings_wndobj", "chrome,resizable");
    wnd.focus();
  },


  initToolbarCmd: function (aID)
  {
    var isToolbarOpen = this._clippingsToolbar.isOpen();
    var toolbarCmd = document.getElementById(aID);
    toolbarCmd.setAttribute("checked", isToolbarOpen);
  },


  toggleClippingsToolbar: function ()
  {
    if (this._clippingsToolbar.isOpen()) {
      this._clippingsToolbar.close();
    }
    else {
      var toolbarWnd = this._clippingsToolbar.open();
      toolbarWnd.focus();
    }
  },


  initClippingsPopup: function (aPopup, aMenu) 
  {
    var err = false;
    var dsURL = this.appUtils.getDataSourcePathURL() + this.appUtils.CLIPDAT_FILE_NAME;
    try {
      var ds = this.clippingsSvc.getDataSource(dsURL);
    }
    catch (e if e.result == Components.results.NS_ERROR_OUT_OF_MEMORY) {
      err = this.strBundle.getString("errorOutOfMemory");
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_ACCESS_DENIED) {
      err = String.ae_fmt("%s: %s",
			  this.strBundle.getString("errorAccessDenied"),
			  this.appUtils.CLIPDAT_FILE_NAME);
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_IS_LOCKED) {
      err = String.ae_fmt("%s: %s",
			  this.strBundle.getString("errorFileLocked"),
			  this.appUtils.CLIPDAT_FILE_NAME);
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_TOO_BIG) {
      err = String.ae_fmt("%s: %s",
			  this.strBundle.getString("errorFileTooBig"),
			  this.appUtils.CLIPDAT_FILE_NAME);
    }
    catch (e) {
      // File is corrupt - open Clippings Manager and perform recovery.
      err = 888;
    }

    var errorMenuItem = document.getElementById("ae-clippings-error-menuitem");

    if (err) {
      // Append a new menu item that lets user see more details of error.
      if (errorMenuItem) {
	return;
      }

      var errorCmd = document.createElement("menuitem");
      errorCmd.id = "ae-clippings-error-menuitem";
      errorCmd.setAttribute("label", this.strBundle.getString("errorMenu"));
      errorCmd.style.fontWeight = "bold";

      var func;
      if (err == 888) {
	func = function () { gClippings.openClippingsManager() };
      }
      else {
	func = function () {
	  gClippings.appUtils.alertEx(gClippings.strBundle.getString('appName'), err);
	};
      }
      errorCmd.addEventListener("command", func, false);
      aPopup.appendChild(errorCmd);
      this._isErrMenuItemVisible = true;
      return;
    }
    else {
      // Remove error menu item if error condition no longer exists
      if (errorMenuItem) {
	aPopup.removeChild(errorMenuItem);
	this._isErrMenuItemVisible = false;
      }
    }

    aMenu.database.AddDataSource(ds);
    aMenu.builder.rebuild();

    this._ds = ds;
    this.dataSrcInitialized = true;
  },


  insertClippingText: function (aURI, aName, aText) 
  {
    var parentFolderURI = this.clippingsSvc.getParent(aURI);
    var folderName = this.clippingsSvc.getName(parentFolderURI);
    var clippingInfo = new aeClippingInfo(aURI, aName, aText, folderName);
    var clippingText = this._clippingSubst.processClippingText(clippingInfo);
    var useClipboard = this.appUtils.getPref("clippings.use_clipboard");

    if (useClipboard) {
      this.appUtils.copyTextToClipboard(clippingText);
      window.setTimeout("gClippings._pasteClipping()", 8);
      return;
    }

    // This will be null if inside a rich edit box
    var textbox = document.commandDispatcher.focusedElement;
    
    if (textbox instanceof HTMLInputElement || textbox instanceof HTMLTextAreaElement) {
      var text, pre, post, pos;
      text = textbox.value;

      if (textbox.selectionStart == textbox.selectionEnd) {
	var point = textbox.selectionStart;
	pre = text.substring(0, point);
	post = text.substring(point, text.length);
	pos = point + clippingText.length;
      }
      else {
	var p1 = textbox.selectionStart;
	var p2 = textbox.selectionEnd;
	pre = text.substring(0, p1);
	post = text.substring(p2, text.length);
	pos = p1 + clippingText.length;
      }

      // "Paste" the clipping, then adjust cursor position so that it is placed
      // after the last character in the "pasted" clipping text
      textbox.value = pre + clippingText + post;
      textbox.selectionStart = pos;
      textbox.selectionEnd = pos;
    }
    else {
      var doc = document.commandDispatcher.focusedWindow.document;
      var hasHTMLTags = clippingText.search(/<[a-z1-6]+( [a-z]+(\="?.*"?)?)*>/i) != -1;
      var hasRestrictedHTMLTags = clippingText.search(/<\?|<%|<!DOCTYPE|(<\b(html|head|body|meta|script|applet|embed|object|i?frame|frameset)\b)/i) != -1;

      if (hasHTMLTags) {
	var pasteAsRichText;

	if (! hasRestrictedHTMLTags) {
	  var showHTMLPasteOpts = this.appUtils.getPref("clippings.html_paste");
	  if (showHTMLPasteOpts == this.HTMLPASTE_ASK_THE_USER) {
	    var dlgArgs = { userCancel: null, pasteAsRichText: null };
	    window.openDialog("chrome://clippings/content/htmlClipping.xul", "htmlClipping_dlg", "chrome,modal,centerscreen", dlgArgs);

	    if (dlgArgs.userCancel) {
	      return;
	    }
	    pasteAsRichText = dlgArgs.pasteAsRichText;
	  }
	  else {
	    pasteAsRichText = showHTMLPasteOpts == this.HTMLPASTE_AS_HTML;
	  }
	}

	if (!pasteAsRichText || hasRestrictedHTMLTags) {
	  clippingText = clippingText.replace(/&/g, "&amp;");
	  clippingText = clippingText.replace(/</g, "&lt;");
	  clippingText = clippingText.replace(/>/g, "&gt;");
	}
      }
      else {
	// Could be plain text but with angle brackets, e.g. for denoting URLs
	// or email addresses, e.g. <joel_user@acme.com>, <http://www.acme.com>
	var hasOpenAngleBrackets = clippingText.search(/</) != -1;
	var hasCloseAngleBrackets = clippingText.search(/>/) != -1;

	if (hasOpenAngleBrackets) {
	  clippingText = clippingText.replace(/</g, "&lt;");
	}
	if (hasCloseAngleBrackets) {
	  clippingText = clippingText.replace(/>/g, "&gt;");	  
	}
      }

      var autoLineBreak = this.appUtils.getPref("clippings.html_auto_line_break");
      var hasLineBreakTags = clippingText.search(/<br|<p/i) != -1;
      if (autoLineBreak && !hasLineBreakTags) {
	clippingText = clippingText.replace(/\n/g, "<br>");
      }

      try {
	doc.execCommand("insertHTML", false, clippingText);
      }
      catch (e) {}
    }
  },


  _pasteClipping: function ()
  {
    try {
      // Paste clipping.  The following function is defined in
      // "chrome://global/content/globalOverlay.js"
      goDoCommand('cmd_paste');
      // SIDE EFFECT: The clipping text will remain on the system clipboard.
    }
    catch (e) {
      // Exception thrown if command is disabled or not applicable
      var showFailMsg = gClippings.appUtils.getPref("clippings.warn_paste_failure");
      if (showFailMsg) {
	gClippings.alert("Clippings paste failure:\n\n" + e);
      }
    }
  },


  keyboardInsertClipping: function (aEvent)
  {
    if (! this._isContentAreaTextBoxFocused()) {
      this.appUtils.beep();
      return;
    } 

    if (this.appUtils.isCommonDataSourceEnabled()) {
      var clippingsMenu1 = document.getElementById("ae-clippings-menu-1");
      clippingsMenu1.builder.refresh();
      clippingsMenu1.builder.rebuild();
    }

    var dlgArgs = { key: null };
    var dlg = window.openDialog("chrome://clippings/content/clippingKey.xul",
				"clipkey_dlg", "modal,centerscreen", dlgArgs);

    this.appUtils.log("keyboardInsertClipping(): char code of key: `" + dlgArgs.key + "'");

    if (dlgArgs.key && dlgArgs.key != aEvent.DOM_VK_ESCAPE) {
      var key = String.fromCharCode(dlgArgs.key);
      var keyDict = this.clippingsSvc.getShortcutKeyDict();

      if (! keyDict.hasKey(key)) {
	this.appUtils.beep();
	return;
      }

      try {
	var valueStr = keyDict.getValue(key);
      }
      catch (e) {}

      valueStr = valueStr.QueryInterface(Components.interfaces.nsISupportsString);
      var clippingURI = valueStr.data;
      this.insertClippingText(clippingURI,
			      this.clippingsSvc.getName(clippingURI),
			      this.clippingsSvc.getText(clippingURI));
    }
  },


  _isContentAreaTextBoxFocused: function ()
  {
    var rv = false;
    var focusedElt = document.commandDispatcher.focusedElement;

    if (focusedElt instanceof HTMLInputElement || focusedElt instanceof HTMLTextAreaElement) {
      // <input type="text"> or <textarea>
      rv = true;
    }
    else {
      // Rich edit box - an <iframe> with designMode == "on"
      var doc = document.commandDispatcher.focusedWindow.document;
      rv = doc.designMode == "on";
    }

    return rv;
  },


  saveClippings: function () 
  {
    var title = this.strBundle.getString('appName');
    try {
      this.clippingsSvc.flushDataSrc();
    }
    catch (e if e.result == Components.results.NS_ERROR_NOT_INITIALIZED) {
      this.appUtils.alertEx(title, this.strBundle.getString("errorSaveFailedDSNotInitialized"));
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_OUT_OF_MEMORY) {
      this.appUtils.alertEx(title, this.strBundle.getString("errorOutOfMemory"));
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_ACCESS_DENIED) {
      var msg = String.ae_fmt("%s: %s",
			      this.strBundle.getString("errorAccessDenied"),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_IS_LOCKED) {
      var msg = String.ae_fmt("%s: %s",
			      this.strBundle.getString("errorFileLocked"),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_TOO_BIG) {
      var msg = String.ae_fmt("%s: %s",
			      this.strBundle.getString("errorFileTooBig"),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_READ_ONLY) {
      var msg = String.ae_fmt("%s: %s",
			      this.strBundle.getString('errorFileReadOnly'),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_DISK_FULL) {
      var msg = String.ae_fmt("%s: %s",
			      this.strBundle.getString('errorDiskFull'),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e) {
      this.appUtils.alertEx(title, this.strBundle.getString("alertSaveFailed"));
      return;
    }
  },


  //
  // Browser window and Clippings menu initialization
  //

  isClippingsInitialized: false,

  initClippings: function ()
  {   
    // Workaround to this init function being called multiple times
    if (this.isClippingsInitialized) {
      return;
    }

    this.appUtils = new aeClippingsUtils();
    this.strBundle = document.getElementById("ae-clippings-strings");
    this._clippingSubst = new aeClippingSubst(this.strBundle);

    try {
      this.clippingsSvc = Components.classes["clippings@mozdev.org/clippings;1"].getService(Components.interfaces.nsIClippingsService);
    }
    catch (e) {
      this.alert(e);
    }

    // String used for empty (null) clippings
    this.clippingsSvc.setEmptyClippingString(this.strBundle.getString('emptyClippingLabel'));

    // First-run initialization - import from Clippings 1.x if necessary.
    if (this.appUtils.getPref("clippings.first_run") == true) {
      this._firstRunInit();
      this.appUtils.setPref("clippings.first_run", false);
    }

    // First-run initialization after upgrade from 1.x/2.x -> 3.0
    if (this.appUtils.getPref("clippings.v3.first_run") == true) {
      var clippings3 = new aeClippings3(this.clippingsSvc, this.strBundle);
      var initFinished = clippings3.startInit();
      if (initFinished) {
	this.appUtils.setPref("clippings.v3.first_run", false);
      }
    }

    var commonDSVerifyMsg = "";
    if (this.appUtils.getPref("clippings.datasource.common")) {
      commonDSVerifyMsg = "Enabled common data source between Firefox and Thunderbird.";
    }

    // Set up Clippings backup.
    var backupDirURL = this.appUtils.getDataSourcePathURL() + this.appUtils.BACKUP_DIR_NAME;
    this.clippingsSvc.setBackupDir(backupDirURL);
    this.clippingsSvc.setMaxBackupFiles(this.appUtils.getPref('clippings.backup.maxfiles'));

    // Initializing data source on Clippings context menu
    var menu = document.getElementById("ae-clippings-menu-1");
    var popup = document.getElementById("ae-clippings-popup-1");
    this.initClippingsPopup(popup, menu);

    this.appUtils.log(String.ae_fmt("gClippings.initClippings(): Initializing Clippings integration with host app window\nHost app: %s (version %s)\n%s", this.appUtils.getHostAppName(), this.appUtils.getHostAppVersionStr(), commonDSVerifyMsg));

    // Add null clipping to root folder if there are no items
    if (this.appUtils.getPref("clippings.datasource.process_root") == true) {
      this.clippingsSvc.processRootFolder();
      this.appUtils.setPref("clippings.datasource.process_root", false);
    }

    // Attaching event handlers to browser context menu 
    var hostAppContextMenu = document.getElementById("contentAreaContextMenu");

    hostAppContextMenu.addEventListener("popupshowing", 
					gClippings._fnInitContextMenuItem, 
					false);
    hostAppContextMenu.addEventListener("popuphiding",
					gClippings._fnHideContextMenuItem,
					false);

    this._clippingsListener = {
      origin:  this.clippingsSvc.ORIGIN_HOSTAPP,

      newFolderCreated: function (aFolderURI) {
	this._rebuildMenu();
      },

      newClippingCreated: function (aClippingURI) {},

      importDone: function (aNumItems) {
	this._rebuildMenu();
      },

      migrationDone: function (aDataSrcURL) {
	var menu = document.getElementById("ae-clippings-menu-1");
	var popup = document.getElementById("ae-clippings-popup-1");

	// Reinitialize Clippings menu so that it points to the correct datasrc
	menu.database.RemoveDataSource(gClippings._ds);
	gClippings.initClippingsPopup(popup, menu);
      },

      _rebuildMenu: function () {
	gClippings.appUtils.log("Rebuilding Clippings submenu");
	var menu = document.getElementById("ae-clippings-menu-1");
	menu.builder.rebuild();
	gClippings.appUtils.log("** Clippings menu rebuild completed successfully **");
      }
    };
    this.clippingsSvc.addListener(this._clippingsListener);

    // Set behaviour of "New Clipping" commands - prompt vs. silent operation
    this.showDialog = true;
    try {
      this.showDialog = !(this.appUtils.getPref("clippings.entries.add_silently"));
    }
    catch (e) {}

    // Initialize "New From Clipboard" command on status bar icon menu.
    var ellipsis = this.showDialog ? this.strBundle.getString("ellipsis") : "";
    var newFromClpbdCmd = document.getElementById("ae_new_clipping_from_clpbd");
    newFromClpbdCmd.setAttribute("label",
				 this.strBundle.getString("newFromClipbd")
				 + ellipsis);

    // Initialize Clippings toolbar
    var toolbarEnabled = false;
    try {
      toolbarEnabled = this.appUtils.getPref("clippings.toolbar.enabled");
    }
    catch (e) {}

    if (toolbarEnabled) {
      this._clippingsToolbar = Components.classes["clippings@mozdev.org/clippings-toolbar;1"].getService(Components.interfaces.nsIClippingsToolbarService);
      this._clippingsToolbar.init(this.clippingsSvc);
      document.getElementById("ae-clippings-toggle-toolbar-2").removeAttribute("hidden");
      document.getElementById("ae-clippings-statusmenu-separator-1").removeAttribute("hidden");
    }
    else {
      this.focusHostAppWnd = function () {};
      this.initToolbarCmd = function () {};
    }

    this.isClippingsInitialized = true;
  },

  // Next two methods are event handlers for the Clippings popup menu; they are
  // to be invoked by the event handling code only.
  _fnInitContextMenuItem: function (aEvent) {
    gClippings.initContextMenuItem.apply(gClippings, arguments);
  },

  _fnHideContextMenuItem: function (aEvent) {
    gClippings.hideContextMenuItem.apply(gClippings, arguments);
  },


  _firstRunInit: function ()
  {
    var oldClipdatFile, clipdat2File;  // nsIFile objects
    oldClipdatFile = this.appUtils.getCurrentProfileDir();
    clipdat2File = this.appUtils.getCurrentProfileDir();
    clipdat2File.append(this.appUtils.CLIPDAT_FILE_NAME);
    oldClipdatFile.append(this.appUtils.CLIPDAT_1X_FILE_NAME);
    
    if (!oldClipdatFile.exists() || !oldClipdatFile.isFile()) {
      this.appUtils.log("It appears that Clippings 1.x was never installed previously.  Welcome!");
      return;
    }
    if (clipdat2File.exists() && clipdat2File.isFile()) {
      this.appUtils.log("Clippings data source file already exists.  Skipping first-run initialization.");
      return;
    }

    this.initClippingsPopup(document.getElementById("ae-clippings-popup-1"),
			    document.getElementById("ae-clippings-menu-1"));
    var numImported = -1;
    var oldClipdatURL = this.appUtils.getDataSourcePathURL() + this.appUtils.CLIPDAT_1X_FILE_NAME;
    try {
      numImported = this.clippingsSvc.importFromFile(oldClipdatURL, true, false, {});
    }
    catch (e) {
      this.appUtils.alertEx(this.strBundle.getString("appName"),
			    this.strBundle.getString("errorImportFailed"));
      this.appUtils.log(e);
    }

    if (numImported != -1) {
      this.appUtils.log("Your Clippings 1.x data was imported successfully.");

      try {
	this.clippingsSvc.flushDataSrc();
	this.appUtils.log("Flushed imported data to disk.");
      }
      catch (e) {
	this.alert(this.strBundle.getString("alertSaveFailed") + "\n" + e);
      }
    }
  },


  initContextMenuItem: function (aEvent)
  {
    // This method performs initialization of menu items of the Clippings
    // popup menus - the one that appears in a text box or rich text box
    // (id='ae-clippings-popup-1') and the one for the browser content area
    // (id='ae-clippings-popop-2').

    if (! this.dataSrcInitialized) {
      throw "gClippings.initContextMenuItem(): Data source not initialized!";
    }

    if (aEvent.target.id != "contentAreaContextMenu") {
      this.appUtils.log("gClippings.initContextMenuItem(): event.target.id is not 'contentAreaContextMenu'; returning");
      return;
    }

    if (! gContextMenu) {
      this.appUtils.log("gClippings.initContextMenuItem(): method invoked by unknown caller; gContextMenu object is null; returning");
      return;
    }

    var clippingsMenu1 = document.getElementById("ae-clippings-menu-1");
    var popup = document.getElementById("ae-clippings-popup-1");

    // Reattach the Clippings datasource to the Clippings menu if it is lost.
    // This is known to occur on Firefox 3 if the Menu Editor extension is
    // installed and its datasource is loaded (menuedit.rdf).
    var dataSrcs = clippingsMenu1.database.GetDataSources();
    var dataSrcFound = false;
    var re = new RegExp(this.appUtils.CLIPDAT_FILE_NAME + "$");   

    while (!dataSrcFound && dataSrcs.hasMoreElements()) {
      var dataSrc = dataSrcs.getNext();
      dataSrc = dataSrc.QueryInterface(Components.interfaces.nsIRDFDataSource);
      if (re.test(dataSrc.URI)) {
	dataSrcFound = true;
      }
    }

    if (! dataSrcFound) {
      this.appUtils.log("gClippings.initContextMenuItem(): Clippings datasource is missing from the Clippings menu's datasources\nInvoking gClippings.initClippingsPopup() to reattach it.");
      this.initClippingsPopup(popup, clippingsMenu1);
    }

    if (this.appUtils.isCommonDataSourceEnabled()) {
      this.appUtils.log("gClippings.initContextMenuItem(): Common datasource enabled; refreshing menu");
      clippingsMenu1.builder.refresh();
    }

    this.appUtils.log("gClippings.initContextMenuItem(): Rebuilding menu");
    clippingsMenu1.builder.rebuild();

    var menu = null;
    var strBundle = document.getElementById("ae-clippings-strings");
    var ellipsis = this.showDialog ? this.strBundle.getString("ellipsis") : "";
    var clippingsMenu2 = document.getElementById("ae-clippings-menu-2");

    var addEntryCmd = document.getElementById("ae_new_clipping_from_textbox");

    // ** FIREFOX 2 **
    // gContextMenu.onTextInput is also true inside a rich edit box!
    if (gContextMenu.onTextInput) {
      var textbox = document.popupNode;

      // Do our own checking to ensure that the node is an HTML text field.
      if (textbox instanceof HTMLInputElement || textbox instanceof HTMLTextAreaElement) {
	if (textbox.selectionStart == textbox.selectionEnd) {
	  // No text selected.  Enable "New..." cmd if text field non-empty
	  addEntryCmd.setAttribute("label",
				   this.strBundle.getString("new") + ellipsis);
	  addEntryCmd.setAttribute("disabled", textbox.value == "");
	}
	else {
	  // "New From Selection..." command.
	  addEntryCmd.setAttribute("label",
				   this.strBundle.getString("newFromSelect") 
				   + ellipsis);
	  addEntryCmd.setAttribute("disabled", "false");
	}
      }
      else {
	var doc = document.popupNode.ownerDocument;

	// ** FIREFOX 2 **  Rich edit box.
	if (this.appUtils.getHostAppVersionStr().split(".")[0] >= 2 && doc) {
	  if (gContextMenu.isTextSelected) {
	    clippingsMenu1.hidden = false;
	    clippingsMenu2.hidden = true;
	    menu = clippingsMenu1;
	      
	    addEntryCmd.setAttribute("disabled", "false");
	    addEntryCmd.setAttribute("label", this.strBundle.getString("newFromSelect") + ellipsis);
	  }
	  else {
	    // Check for empty document
	    var range = doc.createRange();
	    range.setStart(doc.body.firstChild, 0);
	    range.setEnd(doc.body.lastChild, 0);
	    addEntryCmd.setAttribute("disabled", range == "");

	    clippingsMenu1.hidden = false;
	    clippingsMenu2.hidden = true;
	    menu = clippingsMenu1;
	    addEntryCmd.setAttribute("label", this.strBundle.getString("new") + ellipsis);
	  }
	}
	else {
	  addEntryCmd.setAttribute("disabled", "true");
	}
      }
      clippingsMenu1.hidden = false;
      clippingsMenu2.hidden = true;
      menu = clippingsMenu1;
    }
    // Selected text in browser content area or in an <IFRAME>
    else if (gContextMenu.isTextSelected) {
      var doc = document.popupNode.ownerDocument;

      // ** FIREFOX 1.x ONLY **
      // Rich edit box
      if (doc.designMode == "on") {
	clippingsMenu1.hidden = false;
	clippingsMenu2.hidden = true;
	menu = clippingsMenu1;

	addEntryCmd.setAttribute("disabled", "false");
	addEntryCmd.setAttribute("label",
				 this.strBundle.getString("newFromSelect") 
				 + ellipsis);
      } 
      // Browser content area.
      else {
	clippingsMenu1.hidden = true;
	clippingsMenu2.hidden = false;

	addEntryCmd = document.getElementById("ae_new_clipping_from_selection");
	addEntryCmd.setAttribute("label",
				 this.strBundle.getString("newFromSelect") 
				 + ellipsis);
      }
    }
    // ** FIREFOX 1.x ONLY **
    // Rich edit box - if there is any text selected, the preceding
    // "else if ..." statement will be executed.
    else if (gContextMenu.inFrame) {
      var doc = document.popupNode.ownerDocument;
      if (doc.designMode == "on") {
	// Check for empty document
	var range = doc.createRange();
	range.setStart(doc.body.firstChild, 0);
	range.setEnd(doc.body.lastChild, 0);
	addEntryCmd.setAttribute("disabled", range == "");

	clippingsMenu1.hidden = false;
	clippingsMenu2.hidden = true;
	menu = clippingsMenu1;
	addEntryCmd.setAttribute("label",
				 this.strBundle.getString("new") + ellipsis);
      }
    }
  },

  hideContextMenuItem: function (aEvent)
  {
    var popup = document.getElementById("ae-clippings-popup-1");
    var clippingsMenu1 = document.getElementById("ae-clippings-menu-1");
    var clippingsMenu2 = document.getElementById("ae-clippings-menu-2");

    if (aEvent.target.id == "contentAreaContextMenu") {
      clippingsMenu1.hidden = true;
      clippingsMenu2.hidden = true;
    }
  },


  unload: function ()
  {
    this.clippingsSvc.removeListener(this._clippingsListener);
    this._clippingsListener = null;
  },


  focusHostAppWnd :  function ()
  {
    if (! gClippings.appUtils) return;

    var ww = Components.classes["@mozilla.org/embedcomp/window-watcher;1"]
                       .getService(Components.interfaces.nsIWindowWatcher);

    var wnd = ww.activeWindow;
    if (wnd instanceof ChromeWindow) {
      if (this._clippingsToolbar.isHidden()) {
	this._clippingsToolbar.open();
	this._clippingsToolbar.setHidden(false);
      }
    }
  }
};


//
// Event handler initialization
//

window.addEventListener("load",   gClippings, false);
window.addEventListener("unload", gClippings, false);
window.addEventListener("focus",  gClippings, false);
