/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Clippings.
 *
 * The Initial Developer of the Original Code is 
 * Alex Eng <ateng@users.sourceforge.net>.
 * Portions created by the Initial Developer are Copyright (C) 2005-2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */

//
// Integration with host application
//


var gClippings = {
  // Constants for HTML paste options
  HTMLPASTE_ASK_THE_USER: 0,
  HTMLPASTE_AS_HTML:      1,
  HTMLPASTE_AS_IS:        2,

  dataSrcInitialized:    false,
  showDialog:            true,
  showPasteOpts:         false,
  subjectLineCxtMenu:    false,
  appUtils:              null,
  clippingsSvc:          null,
  strBundle:             null,
  _clippingsListener:    null,
  _clippingsToolbar:     null,
  _isErrMenuItemVisible: false,
  _clippingSubst:        null,
  _ds:                   null,
  _initContextMenuItem:  null,

  statusBarObserver: {
    onDragOver: function (aEvent, aFlavour, aSession)
    {
      aEvent.target.setAttribute("ae-dnd", "true");
    },

    onDragExit: function (aEvent, aSession)
    {
      aEvent.target.removeAttribute("ae-dnd");
    },

    onDrop: function (aEvent, aTransferData, aSession)
    {
      if (! gClippings.dataSrcInitialized) {
	// The RDF data source has to be initialized if it has not already been
	// done so, otherwise RDF node creation will fail, and the new
	// Clippings entry will never be created.
	// This initialization is done in the code for the Clippings popup
	// menu's `onpopupshowing' event handler.
	gClippings.initClippingsPopup(document.getElementById("ae-clippings-popup-1"), document.getElementById("ae-clippings-menu-1"));
      }

      var text = aTransferData.data;
      var result = gClippings.clippingsSvc.newClippingFromText(text, gClippings.showDialog, window, null, false);
      if (result) {
	window.setTimeout("gClippings.saveClippings();", 100);
      }
    },

    getSupportedFlavours: function ()
    {
      var flavours = new FlavourSet();
      flavours.appendFlavour("text/unicode");
      return flavours;
    }
  },


  // Method handleEvent() effectively makes gClippings an implementation of
  // the EventListener interface; therefore it can be passed as the listener
  // argument to window.addEventListener() and window.removeEventListener()

  handleEvent: function (aEvent)
  {
    // When this method is invoked, 'this' will not refer to gClippings
    var _this = gClippings;

    if (aEvent.type == "load") {
      _this.initClippings();
    }
    else if (aEvent.type == "unload") {
      _this.unload();
      window.removeEventListener("load", _this, false);
      window.removeEventListener("unload", _this, false);
      window.removeEventListener("focus", _this, false);

      var hostAppCxtMenu = document.getElementById("msgComposeContext");
      hostAppCxtMenu.removeEventListener("popupshowing", 
					 _this._fnInitContextMenuItem,
					 false);

      var subjectLineTextbox = document.getElementById("msgSubject");
      var subjectLineHbox = document.getAnonymousElementByAttribute(subjectLineTextbox, "class", "textbox-input-box");
      if (subjectLineHbox) {
	var subjectLineCxtMenuPopup = document.getAnonymousElementByAttribute(subjectLineHbox, "anonid", "input-box-contextmenu");
	if (subjectLineCxtMenuPopup) {
	  subjectLineCxtMenuPopup.removeEventListener("popupshowing",
						      _this._fnInitContextMenuItem,
						      false);
	}
      }
    }
    else if (aEvent.type == "focus") {
      _this.focusHostAppWnd();
    }
  },


  //
  // Methods invoked by overlay code
  //

  alert: function (aMessage)
  {
    var title = this.strBundle.getString('appName');
    this.appUtils.alertEx(title, aMessage);
  },


  newFromClipboard: function () 
  {
    if (! this.dataSrcInitialized) {
      // The RDF data source has to be initialized if it has not already been
      // done so, otherwise RDF node creation will fail, and the new Clippings
      // entry will never be created.
      // This initialization is done in the code for the Clippings popup menu's
      // `onpopupshowing' event handler.
      this.initClippingsPopup(document.getElementById("ae-clippings-popup-1"),
			      document.getElementById("ae-clippings-menu-1"));
    }

    var txt = this.appUtils.getTextFromClipboard();
    if (! txt) {
      this.alert(this.strBundle.getString('errorNoClipboardContent'));
      return;
    }

    var result = this.clippingsSvc.newClippingFromText(txt, this.showDialog, window, null, false);
    if (result) {
      window.setTimeout("gClippings.saveClippings();", 1);
    }
  },


  getSelectedText: function ()
  {
    var rv;
    var focusedWnd = document.commandDispatcher.focusedWindow;
    var focusedElt = document.commandDispatcher.focusedElement;

    this.appUtils.log("gClippings.getSelectedText(): focusedElt = " + focusedElt);
    if (focusedElt instanceof HTMLInputElement) {
      // Subject line text box
      rv = focusedElt.value.substring(focusedElt.selectionStart, focusedElt.selectionEnd);
    }
    else {
      rv = focusedWnd.getSelection().toString();
    }

    return rv;
  },


  extNewClipping: function ()
  {
    // This function is meant to be invoked by the Clippings toolbar
    var cmdCopy = document.getElementById("cmd_copy");
    if (cmdCopy.getAttribute("disabled")) {
      this.appUtils.beep();
      return;
    }

    goDoCommand("cmd_copy");
    var selection = this.appUtils.getTextFromClipboard();

    if (selection) {
      var result = this.clippingsSvc.newClippingFromText(selection, this.showDialog, window, null, false);
      if (result) {
	window.setTimeout("gClippings.saveClippings();", 1);
      }
    }
  },


  newFromSelection: function ()
  {
    if (! this.dataSrcInitialized) {
      this.initClippingsPopup(document.getElementById("ae-clippings-popup-1"),
			      document.getElementById("ae-clippings-menu-1"));
    }

    // Must explicitly close the message compose context menu - otherwise it
    // will reappear while the New Clipping dialog is open if the Clippings 
    // submenu needs to be rebuilt.
    var cxtMenu = document.getElementById("msgComposeContext");
    cxtMenu.hidePopup();

    var selection = this.getSelectedText();
    if (selection) {
      var result = this.clippingsSvc.newClippingFromText(selection, this.showDialog, window, null, false);
      if (result) {
	window.setTimeout("gClippings.saveClippings();", 1);
      }
    }
    else {
      var strBundle = document.getElementById("ae-clippings-strings");
      this.alert(this.strBundle.getString('errorNoSelection'));
    }
  },


  openClippingsManager: function () 
  {
    var wnd = window.open("chrome://clippings/content/clippings.xul",
			  "ae_clippings_wndobj", "chrome,resizable");
    wnd.focus();
  },


  initToolbarCmd: function (aID)
  {
    var isToolbarOpen = this._clippingsToolbar.isOpen();
    var toolbarCmd = document.getElementById(aID);
    toolbarCmd.setAttribute("checked", isToolbarOpen);
  },


  toggleClippingsToolbar: function ()
  {
    if (this._clippingsToolbar.isOpen()) {
      this._clippingsToolbar.close();
    }
    else {
      var toolbarWnd = this._clippingsToolbar.open();
      toolbarWnd.focus();
    }
  },


  toggleShowPasteOptions: function ()
  {
    this.showPasteOpts = !this.showPasteOpts;
    document.getElementById("ae_clippings_show_paste_options").setAttribute("checked", this.showPasteOpts);
  },


  initClippingsPopup: function (aPopup, aMenu) 
  {
    var err = false;
    var dsURL = this.appUtils.getDataSourcePathURL() + this.appUtils.CLIPDAT_FILE_NAME;
    try {
      this._ds = this.clippingsSvc.getDataSource(dsURL);
    }
    catch (e if e.result == Components.results.NS_ERROR_OUT_OF_MEMORY) {
      err = this.strBundle.getString("errorOutOfMemory");
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_ACCESS_DENIED) {
      err = String.ae_fmt("%s: %s",
			  this.strBundle.getString("errorAccessDenied"),
			  this.appUtils.CLIPDAT_FILE_NAME);
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_IS_LOCKED) {
      err = String.ae_fmt("%s: %s",
			  this.strBundle.getString("errorFileLocked"),
			  this.appUtils.CLIPDAT_FILE_NAME);
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_TOO_BIG) {
      err = String.ae_fmt("%s: %s",
			  this.strBundle.getString("errorFileTooBig"),
			  this.appUtils.CLIPDAT_FILE_NAME);
    }
    catch (e) {
      // File is corrupt - open Clippings Manager and perform recovery.
      err = 888;
    }

    var errorMenuItem = document.getElementById("ae-clippings-error-menuitem");

    if (err) {
      // Append a new menu item that lets user see more details of error.
      if (errorMenuItem) {
	return;
      }

      var errorCmd = document.createElement("menuitem");
      errorCmd.id = "ae-clippings-error-menuitem";
      errorCmd.setAttribute("label", this.strBundle.getString("errorMenu"));
      errorCmd.style.fontWeight = "bold";

      var func;
      if (err == 888) {
	func = function () { gClippings.openClippingsManager() };
      }
      else {
	func = function () {
	  gClippings.appUtils.alertEx(gClippings.strBundle.getString('appName'), err);
	};
      }
      errorCmd.addEventListener("command", func, false);
      aPopup.appendChild(errorCmd);
      this._isErrMenuItemVisible = true;
      return;
    }
    else {
      // Remove error menu item if error condition no longer exists
      if (errorMenuItem) {
	aPopup.removeChild(errorMenuItem);
	this._isErrMenuItemVisible = false;
      }
    }

    aMenu.database.AddDataSource(this._ds);
    aMenu.builder.rebuild();

    this.dataSrcInitialized = true;
    this.appUtils.log("gClippings.initClippingsPopup(): Data source initialization completed.");
  },


  insertClippingText: function (aURI, aName, aText) 
  {
    // Must explicitly close the message compose context menu - otherwise it
    // may reappear while the paste options dialog is open.
    var cxtMenu = document.getElementById("msgComposeContext");
    cxtMenu.hidePopup();

    var parentFolderURI = this.clippingsSvc.getParent(aURI);
    var folderName = this.clippingsSvc.getName(parentFolderURI);
    var clippingInfo = new aeClippingInfo(aURI, aName, aText, folderName);
    var clippingText = this._clippingSubst.processClippingText(clippingInfo);
    var pasteAsQuotation = false;
    var useClipboard = this.appUtils.getPref("clippings.use_clipboard");   

    if (! useClipboard) {
      // Paste clipping into subject line
      var focusedElt = document.commandDispatcher.focusedElement;
      if (focusedElt instanceof HTMLInputElement) {
	var pre, post, pos;
	var textbox = focusedElt;
	var text = textbox.value;

	if (textbox.selectionStart == textbox.selectionEnd) {
	  var point = textbox.selectionStart;
	  pre = text.substring(0, point);
	  post = text.substring(point, text.length);
	  pos = point + clippingText.length;
	}
	else {
	  var p1 = textbox.selectionStart;
	  var p2 = textbox.selectionEnd;
	  pre = text.substring(0, p1);
	  post = text.substring(p2, text.length);
	  pos = p1 + clippingText.length;
	}

	// "Paste" the clipping, then adjust cursor position so that it is placed
	// after the last character in the "pasted" clipping text
	textbox.value = pre + clippingText + post;
	textbox.selectionStart = pos;
	textbox.selectionEnd = pos;
	return;
      }

      // If "Show Options When Pasting" is enabled, ask the user if the
      // clipping should be pasted as normal or quoted text.
      if (this.showPasteOpts) {
	var dlgArgs = { userCancel: null };
	window.openDialog("chrome://clippings/content/pasteOptions.xul", "ae_clippings_pasteopt_dlg", "chrome,centerscreen,modal", dlgArgs);

	if (dlgArgs.userCancel) {
	  return;
	}

	pasteAsQuotation = dlgArgs.pasteOption == 1;
      }

      var contentFrame = document.getElementById("content-frame");
      var editor = contentFrame.getEditor(contentFrame.contentWindow);

      // Composing email in rich text (HTML)
      if (gMsgCompose.composeHTML) {
	var htmlEditor = editor.QueryInterface(Components.interfaces.nsIHTMLEditor);
	var hasHTMLTags = clippingText.search(/<[a-z1-6]+( [a-z]+(\="?.*"?)?)*>/i) != -1;
	var hasRestrictedHTMLTags = clippingText.search(/<\?|<%|<!DOCTYPE|(<\b(html|head|body|meta|script|applet|embed|object|i?frame|frameset)\b)/i) != -1;

	if (hasHTMLTags) {
	  var pasteAsRichText;
	  if (! hasRestrictedHTMLTags) {
	    var showHTMLPasteOpts = this.appUtils.getPref("clippings.html_paste");
	    if (showHTMLPasteOpts == this.HTMLPASTE_ASK_THE_USER) {
	      var dlgArgs = { userCancel: null, pasteAsRichText: null };
	      window.openDialog("chrome://clippings/content/htmlClipping.xul", "htmlClipping_dlg", "chrome,modal,centerscreen", dlgArgs);
	      
	      if (dlgArgs.userCancel) {
		return;
	      }
	      pasteAsRichText = dlgArgs.pasteAsRichText;
	    }
	    else {
	      pasteAsRichText = showHTMLPasteOpts == this.HTMLPASTE_AS_HTML;
	    }
	  }
	  var plainTextClipping = clippingText;
	  
	  if (!pasteAsRichText || hasRestrictedHTMLTags) {
	    clippingText = clippingText.replace(/&/g, "&amp;");
	    clippingText = clippingText.replace(/</g, "&lt;");
	    clippingText = clippingText.replace(/>/g, "&gt;");
	  }
	}
	else {
	  // Could be plain text but with angle brackets, e.g. for denoting URLs
	  // or email addresses, e.g. <joel_user@acme.com>, <http://www.acme.com>
	  var hasOpenAngleBrackets = clippingText.search(/</) != -1;
	  var hasCloseAngleBrackets = clippingText.search(/>/) != -1;

	  if (hasOpenAngleBrackets) {
	    clippingText = clippingText.replace(/</g, "&lt;");
	  }
	  if (hasCloseAngleBrackets) {
	    clippingText = clippingText.replace(/>/g, "&gt;");	  
	  }
	}

	var autoLineBreak = this.appUtils.getPref("clippings.html_auto_line_break");
	var hasLineBreakTags = clippingText.search(/<br|<p/i) != -1;
	if (autoLineBreak && !hasLineBreakTags) {
	  clippingText = clippingText.replace(/\n/g, "<br>");
	}
      }
      else {
	// Composing email without formatting
	var plainTextEditor = editor.QueryInterface(Components.interfaces.nsIPlaintextEditor);
      }

      if (pasteAsQuotation) {
	var mailEditor = editor.QueryInterface(Components.interfaces.nsIEditorMailSupport);
	if (gMsgCompose.composeHTML) {
	  mailEditor.insertAsCitedQuotation(clippingText, "", true);
	}
	else {
	  mailEditor.insertAsQuotation(clippingText);
	  mailEditor.rewrap(true);
	}
	return;
      }
      
      if (gMsgCompose.composeHTML) {
	htmlEditor.insertHTML(clippingText);
      }
      else {
	plainTextEditor.insertText(clippingText);
      }
      
      return;
    }

    this.appUtils.copyTextToClipboard(clippingText);

    if (this.showPasteOpts) {
      var dlgArgs = { userCancel: null };
      window.openDialog("chrome://clippings/content/pasteOptions.xul", "ae_clippings_pasteopt_dlg", "chrome,centerscreen,modal", dlgArgs);

      if (dlgArgs.userCancel) {
	return;
      }

      if (dlgArgs.pasteOption == 1) {
	goDoCommand("cmd_pasteQuote");

	// gMsgCompose defined in <chrome://messenger/content/messengercompose/
	// MsgComposeCommands.js>.  It is an instance of nsIMsgCompose,
	// progid "@mozilla.org/messengercompose/compose;1"
	// Line rewrapping not necessary if composing in HTML format
	if (! gMsgCompose.composeHTML) {
	  goDoCommand("cmd_rewrap");
	}
	return;
      }
    }

    window.setTimeout("gClippings._pasteClipping()", 8);
  },


  _pasteClipping: function ()
  {
    try {
      // Paste clipping.  The following function is defined in
      // "chrome://global/content/globalOverlay.js"
      goDoCommand('cmd_paste');
      // SIDE EFFECT: The clipping text will remain on the system clipboard.
    }
    catch (e) {
      // Exception thrown if command is disabled or not applicable
      this.appUtils.beep();
    }
  },


  keyboardInsertClipping: function (aEvent)
  {
    if (this.appUtils.isCommonDataSourceEnabled()) {
      var clippingsMenu1 = document.getElementById("ae-clippings-menu-1");
      clippingsMenu1.builder.refresh();
      clippingsMenu1.builder.rebuild();
    } 

    var dlgArgs = { key: null };
    var dlg = window.openDialog("chrome://clippings/content/clippingKey.xul",
				"clipkey_dlg", "modal,centerscreen", dlgArgs);

    if (dlgArgs.key && dlgArgs.key != aEvent.DOM_VK_ESCAPE) {
      var key = String.fromCharCode(dlgArgs.key);
      var keyDict = this.clippingsSvc.getShortcutKeyDict();

      if (! keyDict.hasKey(key)) {
	this.appUtils.beep();
	return;
      }

      try {
	var valueStr = keyDict.getValue(key);
      }
      catch (e) {}

      valueStr = valueStr.QueryInterface(Components.interfaces.nsISupportsString);
      var clippingURI = valueStr.data;
      this.insertClippingText(clippingURI,
			      this.clippingsSvc.getName(clippingURI),
			      this.clippingsSvc.getText(clippingURI));
    }
  },


  saveClippings: function () 
  {
    var title = this.strBundle.getString('appName');
    try {
      this.clippingsSvc.flushDataSrc();
    }
    catch (e if e.result == Components.results.NS_ERROR_NOT_INITIALIZED) {
      this.appUtils.alertEx(title, this.strBundle.getString("errorSaveFailedDSNotInitialized"));
      return;
	   }
    catch (e if e.result == Components.results.NS_ERROR_OUT_OF_MEMORY) {
      this.appUtils.alertEx(title, this.strBundle.getString("errorOutOfMemory"));
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_ACCESS_DENIED) {
      var msg = String.ae_fmt("%s: %s",
			      this.strBundle.getString("errorAccessDenied"),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_IS_LOCKED) {
      var msg = String.ae_fmt("%s: %s",
			      this.strBundle.getString("errorFileLocked"),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_TOO_BIG) {
      var msg = String.ae_fmt("%s: %s",
			      this.strBundle.getString("errorFileTooBig"),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_READ_ONLY) {
      var msg = String.ae_fmt("%s: %s",
			      this.strBundle.getString('errorFileReadOnly'),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e if e.result == Components.results.NS_ERROR_FILE_DISK_FULL) {
      var msg = String.ae_fmt("%s: %s",
			      this.strBundle.getString('errorDiskFull'),
			      this.appUtils.CLIPDAT_FILE_NAME);
      this.appUtils.alertEx(title, msg);
      return;
    }
    catch (e) {
      this.appUtils.alertEx(title, this.strBundle.getString("alertSaveFailed"));
      return;
    }
  },


  //
  // Browser window and Clippings menu initialization
  //

  isClippingsInitialized: false,

  initClippings: function ()
  {   
    // Workaround to this init function being called multiple times
    if (this.isClippingsInitialized) {
      return;
    }

    this.appUtils = new aeClippingsUtils();
    this.strBundle = document.getElementById("ae-clippings-strings");
    this._clippingSubst = new aeClippingSubst(this.strBundle);

    this.appUtils.log(">> gClippings.initClippings()");

    try {
      this.clippingsSvc = Components.classes["clippings@mozdev.org/clippings;1"].getService(Components.interfaces.nsIClippingsService);
    }
    catch (e) {
      this.alert(e);
    }

    // String used for empty (null) clippings
    this.clippingsSvc.setEmptyClippingString(this.strBundle.getString('emptyClippingLabel'));

    if (this.appUtils.getPref("clippings.first_run") == true) {
      this.appUtils.setPref("clippings.first_run", false);
    }

    // First-run initialization after upgrade from 2.x -> 3.0
    if (this.appUtils.getPref("clippings.v3.first_run") == true) {
      var clippings3 = new aeClippings3(this.clippingsSvc, this.strBundle);
      var initFinished = clippings3.startInit();
      if (initFinished) {
	this.appUtils.setPref("clippings.v3.first_run", false);
      }
    }

    var commonDSVerifyMsg = "";
    if (this.appUtils.getPref("clippings.datasource.common")) {
      commonDSVerifyMsg = "Enabled common data source between Firefox and Thunderbird.";
    }

    // Clippings backup
    var backupDirURL = this.appUtils.getDataSourcePathURL() + this.appUtils.BACKUP_DIR_NAME;
    this.clippingsSvc.setBackupDir(backupDirURL);
    this.clippingsSvc.setMaxBackupFiles(this.appUtils.getPref('clippings.backup.maxfiles'));

    // Initializing data source on Clippings context menus
    var menu1 = document.getElementById("ae-clippings-menu-1");
    var popup1 = document.getElementById("ae-clippings-popup-1");
    this.initClippingsPopup(popup1, menu1);

    // Message subject line context menu
    var menu2 = document.getElementById("ae-clippings-menu-2");
    var popup2 = document.getElementById("ae-clippings-popup-2");
    this.initClippingsPopup(popup2, menu2);


    this.appUtils.log(String.ae_fmt("gClippings.initClippings(): Initializing Clippings integration with host app window: %s\n%s", window.location.href, commonDSVerifyMsg));

    // Add null clipping to root folder if there are no items
    if (this.appUtils.getPref("clippings.datasource.process_root") == true) {
      this.clippingsSvc.processRootFolder();
      this.appUtils.setPref("clippings.datasource.process_root", false);
    }

    // Attaching event handler to context menu 
    var hostAppContextMenu = document.getElementById("msgComposeContext");
    hostAppContextMenu.addEventListener("popupshowing", 
					this._fnInitContextMenuItem, false);

    this._clippingsListener = {
      origin:  this.clippingsSvc.ORIGIN_HOSTAPP,

      migrationDone: function (aDataSrcURL) {
	var menu = document.getElementById("ae-clippings-menu-1");
	var popup = document.getElementById("ae-clippings-popup-1");

	// Reinitialize Clippings menu so that it points to the correct datasrc
	menu.database.RemoveDataSource(gClippings._ds);
	gClippings.initClippingsPopup(popup, menu);
      },

      newFolderCreated:    function (aFolderURI) {},
      newClippingCreated:  function (aClippingURI) {},
      importDone:          function (aNumItems) {}
    };

    this.clippingsSvc.addListener(this._clippingsListener);

    // Set behaviour of "New Clipping" commands - prompt vs. silent operation
    this.showDialog = true;
    try {
      this.showDialog = !(this.appUtils.getPref("clippings.entries.add_silently"));
    }
    catch (e) {}

    // Initialize "New From Clipboard" command on status bar icon menu.
    var ellipsis = this.showDialog ? this.strBundle.getString("ellipsis") : "";
    var newFromClpbdCmd = document.getElementById("ae_new_clipping_from_clpbd");
    newFromClpbdCmd.setAttribute("label",
				 this.strBundle.getString("newFromClipbd")
				 + ellipsis);

    // Initialize Clippings toolbar
    var toolbarEnabled = false;
    try {
      toolbarEnabled = this.appUtils.getPref("clippings.toolbar.enabled");
    }
    catch (e) {}

    if (toolbarEnabled) {
      this._clippingsToolbar = Components.classes["clippings@mozdev.org/clippings-toolbar;1"].getService(Components.interfaces.nsIClippingsToolbarService);
      this._clippingsToolbar.init(this.clippingsSvc);
      document.getElementById("ae-clippings-toggle-toolbar-2").removeAttribute("hidden");
    }
    else {
      this.focusHostAppWnd = function () {};
      this.initToolbarCmd = function () {};
    }

    if (this.subjectLineCxtMenu) {
      var subjectLineTextbox = document.getElementById("msgSubject");
      /**
	 subjectLineTextbox.addEventListener("click", this._checkSubjectLineCxtMenu, false);
      **/
      this._initSubjectLineCxtMenu(subjectLineTextbox);
    }

    this.isClippingsInitialized = true;

    this.appUtils.log("<< gClippings.initClippings()");
  },


  _checkSubjectLineCxtMenu: function (aEvent)
  {
    // This method is called within an event listener, so 'this' won't refer to
    // gClippings
    var subjectLineClippingsMenu = document.getElementById("ae-clippings-menu-2");

    if (! subjectLineClippingsMenu) {
      // subjectLineClippingsMenu would be null if the Clippings submenu has
      // disappeared from the subject line textbox context menu
      var subjectLineTextbox = document.getElementById("msgSubject");
      gClippings._initSubjectLineCxtMenu.apply(gClippings, [subjectLineTextbox]);
    }
  },


  _initSubjectLineCxtMenu: function (aTextbox) 
  {
    this.appUtils.log(">> gClippings._initSubjectLineCxtMenu()");
    
    // document.getAnonymousElementByAttribute() will only look one level below
    // the element argument, i.e. the 1st parameter
    var subjectLineHbox = document.getAnonymousElementByAttribute(aTextbox, "class", "textbox-input-box");
    if (subjectLineHbox) {
      var subjectLineCxtMenuPopup = document.getAnonymousElementByAttribute(subjectLineHbox, "anonid", "input-box-contextmenu");
      if (subjectLineCxtMenuPopup) {
	this.appUtils.log("gClippings._initSubjectLineCxtMenu(): subjectLineCxtMenuPopup = " + subjectLineCxtMenuPopup + "; node name = " + (subjectLineCxtMenuPopup ? subjectLineCxtMenuPopup.nodeName : "(n/a)"));

	// Add the Clippings menu to subjectLineCxtMenuPopup
	var clippingsMenu2 = document.getElementById("ae-clippings-menu-2");
	this.appUtils.log("gClippings._initSubjectLineCxtMenu(): clippingsMenu2 = " + clippingsMenu2);
	subjectLineCxtMenuPopup.appendChild(clippingsMenu2);
	subjectLineCxtMenuPopup.addEventListener("popupshowing",
						 this._fnInitContextMenuItem,
						 false);
      }
    }
    this.appUtils.log("<< gClippings._initSubjectLineCxtMenu()");
  },


  _fnInitContextMenuItem: function (aEvent) {
    if (aEvent.target.id == "msgComposeContext"
	|| aEvent.target.getAttribute("anonid") == "input-box-contextmenu") {
      gClippings.initContextMenuItem.apply(gClippings, arguments);
    }
  },


  initContextMenuItem: function (aEvent)
  {
    this.appUtils.log("gClippings.initContextMenuItem(): Event target: " + aEvent.target + "; tag name: " + aEvent.target.tagName + "; id = `" + aEvent.target.id + "'");

    var clippingsMenu;

    if (aEvent.target.id == "msgComposeContext") {
      clippingsMenu = document.getElementById("ae-clippings-menu-1");
    }
    else if (aEvent.target.getAttribute("anonid") == "input-box-contextmenu") {
      clippingsMenu = document.getElementById("ae-clippings-menu-2");
    }

    this._initCxtMenuItem(clippingsMenu);
  },


  _initCxtMenuItem: function (aMenupopup)
  {
    this.appUtils.log("gClippings._initCxtMenuItem(): aMenupopup = " + aMenupopup + "; tag name: " + aMenupopup.tagName + "; id = `" + aMenupopup.id + "'");

    if (this.appUtils.isCommonDataSourceEnabled()) {
      aMenupopup.builder.refresh();
    }
    aMenupopup.builder.rebuild();

    var strBundle = document.getElementById("ae-clippings-strings");
    var ellipsis = this.showDialog ? this.strBundle.getString("ellipsis") : "";
    var addEntryCmd = document.getElementById("ae_new_clipping_from_selection");
    var selection;

    if (aMenupopup.id == "ae-clippings-menu-1") {
      selection = this.getSelectedText();
    }
    else if (aMenupopup.id == "ae-clippings-menu-2") {
      var subjectLineTextbox = document.getElementById("msgSubject");
      selection = subjectLineTextbox.value.substring(subjectLineTextbox.selectionStart, subjectLineTextbox.selectionEnd);
    }
  
    addEntryCmd.setAttribute("disabled", selection == "");
    addEntryCmd.setAttribute("label", strBundle.getString("newFromSelect")
			     + ellipsis);
  },


  unload: function ()
  {
    this.clippingsSvc.removeListener(this._clippingsListener);
    this._clippingsListener = null;
  },


  focusHostAppWnd :  function ()
  {
    if (! gClippings.appUtils) return;

    var ww = Components.classes["@mozilla.org/embedcomp/window-watcher;1"]
                       .getService(Components.interfaces.nsIWindowWatcher);

    var wnd = ww.activeWindow;
    if (wnd instanceof ChromeWindow) {
      if (this._clippingsToolbar.isHidden()) {
	this._clippingsToolbar.open();
	this._clippingsToolbar.setHidden(false);
      }
    }
  }
};


//
// Event handler initialization
//

window.addEventListener("load",   gClippings, false);
window.addEventListener("unload", gClippings, false);
window.addEventListener("focus",  gClippings, false);

